import groovy.xml.StreamingMarkupBuilder
import org.xml.sax.helpers.DefaultHandler
import org.xml.sax.Attributes
import org.xml.sax.InputSource
import org.xml.sax.SAXException
import javax.xml.parsers.SAXParserFactory
import java.util.regex.Pattern
import java.util.regex.Matcher
import java.text.SimpleDateFormat

def isWindows = (System.getProperty('os.name') =~ /(?i)windows/).find()
def ahptool = isWindows ? 'ahptool.cmd' : 'ahptool'
def out = System.out

def runCommand = { message, command, Closure closure ->
    ProcessBuilder builder = new ProcessBuilder()  
    builder.command(command as String[])
    println()
    if (message) {
        println(message)
    }
    println("command: ${builder.command().join(' ')}")
    def process = builder.start()
    if (closure) {
        closure(process)
    }
    else {
        process.consumeProcessOutput(out, out) // forward stdout and stderr
        process.out.close() // close stdin
    }
    process.waitFor()
    if (process.exitValue()) {
       throw new Exception("Command failed with exit code: "+process.exitValue())
    }
}

def getProperties = { scopes ->
    if (!(scopes instanceof Collection)) {
         scopes = [scopes];
    }
     
    def props = [:]

    scopes.each { scope ->
        def ahptoolProc = (ahptool + " get${scope}Properties").execute()
        ahptoolProc.consumeProcessErrorStream(System.out);
        def propertiesXml = ahptoolProc.text
        ahptoolProc.waitFor()
        if (ahptoolProc.exitValue()) {
            throw new Exception("Failed to get step properties from ahptool: " + propertiesXml)
        }
        new XmlSlurper().parseText(propertiesXml).property.each{ propElem ->
            props[propElem.@name.text()] = propElem.text()
        }
    }
    return props;
}

class FileSet {
    def isWindows = (System.getProperty('os.name') =~ /(?i)windows/).find()
    def base
    def includes = []
    def excludes = []

    def FileSet(base) {
        if (base instanceof File) {
            this.base = base
        }
        else if (base == null) {
            this.base = new File('.').absoluteFile
        }
        else {
            this.base = new File(base)
        }
    }

    def include(antPattern) {
        forNonEmptyLines(antPattern) {
            includes << convertToPattern(it) 
        }
    }

    def exclude(antPattern) {
        forNonEmptyLines(antPattern) {
            excludes << convertToPattern(it)
        }
    }

    def each(Closure closure) {
        base.eachFileRecurse { file ->
            def path = file.path.replace('\\', '/').substring(base.path.length())
            def matches = false
            for (p in includes) {
                if (path =~ p) {
                    matches = true
                    break;
                }
            }
            if (matches) {
                for (p in excludes) {
                    if (path =~ p) {
                        matches = false;
                        break;
                    }
                }
            }
            if (matches) {
                closure(file)
            }
        }
    }

    def files() {
        def list = []
        each { list << it }
        return list
    }

    private forNonEmptyLines(strings, Closure closure) {
        if (strings instanceof Collection) {
            strings.each { 
                def trimmed = it?.trim()
                if (trimmed?.length() > 0) {
                    closure(it)
                }
            }
        }
        else if (strings != null) {
            forNonEmptyLines(strings.readLines(), closure)
        }
    }

    private convertToPattern(antPattern) {
        // trim leading / character from pattern
        def pattern = antPattern.startsWith('/') ? antPattern : '/'+antPattern

        // deal with special regex-characters that should be interpreted as literals
        '\\.+[]^${}|()'.toCharArray().each{ c ->
            pattern = pattern.replace(''+c, '\\'+c)
        }
        pattern = pattern.replace('?', '.') // ? is a single-char wildcard
        
        // deal with ant-style wildcards
        StringBuffer result = new StringBuffer()
        result.append("^")
        def m = (pattern =~ '\\*\\*/|\\*\\*|\\*')
        while (m) {
            def token = m.group()
            def replacement;
            if (token == '**/') {
                replacement = '.*(?<=/)'
            }
            else if (token == '**') {
                replacement = '.*'
            }
            else {
                replacement = '[^/]*'
            }
            m.appendReplacement(result, Matcher.quoteReplacement(replacement))
        }
        m.appendTail(result)
        result.append("\$")
        def flags = 0
        if (isWindows) {
            flags |= Pattern.CASE_INSENSITIVE
        }
        return Pattern.compile(result.toString(), flags)
    }
}

class MSTestHandler extends DefaultHandler {
    final SUCCESS = "success"
    final FAILURE = "failure"
    final ERROR = "error"

    final millisPerSecond = 1000
    final millisPerMinute = 60 * millisPerSecond
    final millisPerHour = 60 * millisPerMinute

    final SimpleDateFormat vs2010Date = new SimpleDateFormat("yyyy-MM-dd hh:mm:ss.SSS");
    
    final resultMap = buildResultMap()

    def filename
    def onNewTest
    def testTable = [:]

    def inUnitTest
    def inUnitTestResult
    def inMessage

    def testName
    def testResult
    def testDuration
    def testId

    def unitTestId
    def unitTestClassName
    def unitTestName

    def messages = []
    def currentText = new StringBuilder()

    MSTestHandler(def filename) {
        this.filename = filename
    }

    void startElement(String ns, String localName, String qName, Attributes attrs) {
        switch (qName) {
            case 'UnitTest':
                inUnitTest = true
                unitTestId = attrs.getValue('id')
                break;

            case 'UnitTestResult':
                inUnitTestResult = true
                testId = attrs.getValue('testId')
                testName = attrs.getValue('testName')
                testResult = toTestResult(attrs.getValue('outcome'))
                String durationStr;
                durationStr = attrs.getValue('duration')
                if (durationStr == null) {
                    String startTime = attrs.getValue('startTime').replace('T',' ').substring(0,23)
                    String endTime = attrs.getValue('endTime').replace('T', ' ').substring(0,23)
                    def startDate = vs2010Date.parse(startTime)
                    def endDate = vs2010Date.parse(endTime)
                    testDuration = endDate.getTime() - startDate.getTime()
                    
                }
                else {
                    testDuration = toTestDuration(durationStr)

                }
                break;

            case 'Message':
                if (inUnitTestResult) {
                    inMessage = true
                }
                break;

            case 'TestMethod':
                if (inUnitTest) {
                    unitTestClassName = getClassName(attrs.getValue('className'))
                    unitTestName = attrs.getValue('name')
                }
                break;
        }
    }

    void endElement(String ns, String localName, String qName) {
        switch (qName) {
            case 'UnitTest':
                testTable[unitTestId] = [className: unitTestClassName, name: unitTestName]
                inUnitTest = false
                break;

            case 'UnitTestResult':
                def className = testTable[testId].className
                onNewTest(testName, className, testResult, testDuration, messages)
                inUnitTestResult = false
                messages = []
                break;

            case 'Message':
                if (inUnitTestResult) {
                    messages << currentText.toString()
                    currentText = new StringBuilder()
                    inMessage = false;
                }
                break;
        }
    }
    
    void characters(char[] ch, int start, int length) {
        if (inMessage) {
            currentText.append(ch, start, length)
        }
    }
    
    def toTestResult(msTestOutcome) {
        return resultMap[msTestOutcome]
    }

    def toTestDuration(msTestDuration) {
        def matcher = msTestDuration =~ /(\d+):(\d+):(\d+(?:.\d+)?)/
        def hours = matcher[0][1]
        def minutes = matcher[0][2]
        def seconds = matcher[0][3]
        def duration = hours.toInteger() * millisPerHour + 
                       minutes.toInteger() * millisPerMinute + 
                       seconds.toBigDecimal() * millisPerSecond
        return duration.setScale(0, java.math.RoundingMode.HALF_UP)
    }

    def buildResultMap() {
        def map = [
            'Error': ERROR,
            'Failed': FAILURE,
            'Timeout': FAILURE,
            'Aborted': FAILURE,
            'Inconclusive': FAILURE,
            'PassedButRunAborted': SUCCESS,
            'NotRunnable': ERROR,
            'NotExecute': FAILURE,
            'Disconnected': ERROR,
            'Warning': SUCCESS,
            'Passed': SUCCESS,
            'Completed': SUCCESS,
            'InProgress': ERROR,
            'Pending': ERROR]
        return map.asImmutable()
    }

    def getClassName(classStrongName) {
        def i = classStrongName.indexOf(',')
        def className = classStrongName[0 ..< i].trim()
        className = className.replaceAll(/\+/, '.')
        className = className.replaceAll('\\\\\\\\', '\\\\')
        return className
    }

    def readTests(Closure closure) {
        onNewTest = closure
        def reader = SAXParserFactory.newInstance().newSAXParser().XMLReader
        def suiteFile = (filename instanceof File) ? filename : new File(filename)
        suiteFile.withInputStream { stream ->
            reader.setContentHandler(this)
            reader.parse(new InputSource(stream))
        }
    }
}

class EndParseEarlyException extends SAXException {}

class SuiteHandler extends DefaultHandler {
 
    def filename
    def suiteName

    SuiteHandler(filename) {
        this.filename = filename 
    }

    void startElement(String ns, String localName, String qName, Attributes attrs) {
        if (qName == 'TestRunConfiguration') {
            suiteName = attrs.getValue('name')
            throw new EndParseEarlyException()
        }
    }

    def getSuiteName() {
        def reader = SAXParserFactory.newInstance().newSAXParser().XMLReader
        try {
            def suiteFile = (filename instanceof File) ? filename : new File(filename)
            suiteFile.withInputStream { stream ->
                reader.setContentHandler(this)
                reader.parse(new InputSource(stream))
            }
        }
        catch (EndParseEarlyException swallow) {
        }
        return suiteName
    }
}

def properties = getProperties(['Step'])
def reportName = properties.reportName

def fileSet = new FileSet(properties.baseDir)
fileSet.include(properties.includes)
fileSet.exclude(properties.excludes)
def reportFiles = fileSet.files()

def builder = new StreamingMarkupBuilder()
builder.encoding = 'UTF-8'

def xml = builder.bind() {
    mkp.xmlDeclaration()

    'test-report'(name: reportName, type: 'MSTest', 'job-id': System.getenv('AH_JOB_ID')) {
        reportFiles.each { file ->
            println("Parsing " + file.absolutePath)
            def suiteName = new SuiteHandler(file).getSuiteName()
            'test-suite'(name: suiteName) {
                def handler = new MSTestHandler(file)
                handler.readTests { testName, testClass, testResult, testDuration, messages ->
                    test(name: testName, 'class-name': testClass, result: testResult, time: testDuration) {
                        messages.each {
                            message(it)
                        }
                    }
                }
            }
        }
    }
}

runCommand("Uploading to AHP", [ahptool, 'addTests', '-report', reportName, '-']) { process ->
    process.consumeProcessOutput(out, out) // forward process stdout and stderr to our stdout
    process.withWriter { it << xml }
}
